//
// shatbox (C) 2001 Matthew Conte (matt@conte.com)
//
//
// apu.h
//
// apu emulation
//
// $Id: $
//

#ifndef _APU_H_
#define _APU_H_

#include "typedefs.h"
#include "unit.h"


class soundChannel
{
public:
   virtual void reset();
   virtual void write(int reg, uint8 value);
   virtual void phaseChange(int phase);

   virtual int  getSample(int rate) = 0;
   virtual void activate(bool state);
   virtual bool isActive();

protected:
   enum
   {
      DECAY_SHIFT = 5,
      DECAY_DIV = (1 << DECAY_SHIFT),
   };

   int      mOutput;
   int      mOutputTimer;
   int      mTimer;
   uint16   mFreq;
   uint8    mLengthCounter;
   
   uint8    mRegs[4];
   bool     mEnabled;
};

class rectangleChannel : public soundChannel
{
public:
   rectangleChannel(int which);

   virtual void reset();
   virtual void write(int reg, uint8 value);

   virtual int  getSample(int rate);
   virtual void phaseChange(int phase);

private:
   int      mWhich;

   uint8    mStep;
   uint8    mEnvCount;
   uint8    mEnvRate;
   uint8    mSweepRate;
   uint16   mSweepCarry;
   uint8    mDutyPeriod;
};

class triangleChannel : public soundChannel
{
public:
   triangleChannel();

   virtual void reset();
   virtual void write(int reg, uint8 value);

   virtual int  getSample(int rate);
   virtual void phaseChange(int phase);

private:
   uint8    mStep;
   int      mLatency;
   bool     mCounting;
   uint8    mLinearCounter;
};

class noiseChannel : public soundChannel
{
public:
   noiseChannel();

   virtual void reset();
   virtual void write(int reg, uint8 value);

   virtual int  getSample(int rate);
   virtual void phaseChange(int phase);

private:
   uint8    shiftRegister15();

   uint8    mEnvCount;
   uint8    mEnvRate;
   int      mState;
   uint16   mShiftReg;
   uint8    mXorTap;
};

class cpu;
class dmcChannel : public soundChannel
{
public:
   dmcChannel(cpu *pCPU);

   virtual void reset();
   virtual void write(int reg, uint8 value);

   virtual int  getSample(int rate);
   virtual void activate(bool state);
   virtual bool isActive();

   bool  getIRQFlag() const;

private:
   // access to CPU for IRQ/DMA
   cpu      *mpCPU;
   int      mLength;
   uint8    mShiftReg;
   uint32   mAddress;
   bool     mIRQFlag;
   uint8    mOldVol;
};

inline bool dmcChannel::getIRQFlag() const
{
   return mIRQFlag;
}


class apu : public unit
{
public:

   apu(machine *pMachine, cpu *pCPU, int mixingRate);
   virtual ~apu();

   virtual void reset(bool hardReset);

   void  advance(int cpuCycles);
   void  update();

   // returns number of samples available
   int   flushSamples(unsigned char **ppStream);

   static uint8 dispatchReadCtrl(void *userdata, uint32 offset);
   static void  dispatchWrite(void *userdata, uint32 offset, uint8 value);
   static void  dispatchWriteCtrl(void *userdata, uint32 offset, uint8 value);

   enum
   {
      APU_FIX = 10,
   };

private:
   enum
   {
      CHAN_RECT0 = 0,
      CHAN_RECT1,
      CHAN_TRI,
      CHAN_NOISE,
      CHAN_DMC,
      NUM_CHANNELS
   };

   enum
   {
      //APU_CLOCK      = (236250000 / 11),
      APU_CLOCK      = 21442080,
      APU_RATE       = ((APU_CLOCK / 12) << APU_FIX),
   };

   uint8 readCtrl(uint32 offset);
   void  write(uint32 offset, uint8 value);
   void  writeCtrl(uint32 offset, uint8 value);

   soundChannel   *mpChannel[NUM_CHANNELS];

   int   mRate;
   int   mCycles;
   int   mPhaseCount;
   int   mPhase;

   uint8 mControlReg;

   uint8 *mpMixBuffer;
   int   mBufferPos;
   int   mBufferSize;

public:
   enum
   {
      APU_FRAMELEN   = (APU_RATE / 60),
      OUTPUT_FIX     = 16,
   };
};

inline void apu::advance(int cpuCycles)
{
   mCycles -= cpuCycles << APU_FIX;
}

#endif // !_APU_H_
