/*
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
**
**
** config.cpp
**
** Configuration API 
**
** This file was written by Ben Sittler (bsittler@iname.com)
** and is in the public domain.
**
** $Id: $
*/

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "typedefs.h"
#include "assert.h"
#include "config.h"
#include "version"


/* avoid name-clashing on some systems */
static int my_stricmp(const char *s1, const char *s2)
{
   while (tolower((unsigned char) *s1) == tolower((unsigned char) *s2))
   {
      if (0 == *s1)
         return 0;
      s1++;
      s2++;
   }

   return (int) tolower((unsigned char) *s1) - (int) tolower((unsigned char) *s2);
}


static void my_destroy(myvar_t **var)
{
   ASSERT(*var);

   if ((*var)->group) 
      free((*var)->group);
   if ((*var)->key)
      free((*var)->key);
   if ((*var)->value)
      free((*var)->value);
   free(*var);
}

static myvar_t *my_create(const char *group, const char *key, const char *value)
{
   myvar_t *var = (myvar_t *) malloc(sizeof(*var));
   if (NULL == var)
   {
      return 0;
   }

   var->less = var->greater = NULL;
   var->group = var->key = var->value = NULL;

   if ((var->group = (char *) malloc(strlen(group) + 1))
       && (var->key = (char *) malloc(strlen(key) + 1))
       && (var->value = (char *) malloc(strlen(value) + 1)))
   {
      strcpy(var->group, group);
      strcpy(var->key, key);
      strcpy(var->value, value);

      return var;
   }

   my_destroy(&var);
   return NULL;
}

static myvar_t *my_lookup(myvar_t *vars, const char *group, const char *key) 
{
   int cmp;
   myvar_t *current = vars;

   while (current
          && ((cmp = my_stricmp(group, current->group))
              || (cmp = my_stricmp(key, current->key))))
   {
      if (cmp < 0)
         current = current->less;
      else
         current = current->greater;
   }

   return current;
}

static void my_insert(myvar_t **vars, myvar_t *var)
{
   int cmp;
   myvar_t **current = vars;

   while (*current
          && ((cmp = my_stricmp(var->group, (*current)->group))
              || (cmp = my_stricmp(var->key, (*current)->key))))
   {
      current = (cmp < 0) ? &(*current)->less : &(*current)->greater;
   }

   if (*current)
   {
      var->less = (*current)->less;
      var->greater = (*current)->greater;
      my_destroy(current);
   }
   else
   {
      var->less = var->greater = NULL;
   }

   *current = var;
}

static void my_save(FILE *stream, myvar_t *var, char **group)
{
   if (NULL == var)
      return;

   my_save(stream, var->less, group);
   
   if (my_stricmp(*group, var->group))
   {
      fprintf(stream, "\n[%s]\n", var->group);
      *group = var->group;
   }
   
   fprintf(stream, "%s=%s\n", var->key, var->value);
   
   my_save(stream, var->greater, group);
}

static void my_cleanup(myvar_t *var)
{
   if (NULL == var)
      return;

   my_cleanup(var->less);
   my_cleanup(var->greater);
   my_destroy(&var);
}

static char *my_getline(FILE *stream)
{
   char buf[1024];
   char *dynamic = NULL;

   do
   {
      if (NULL == (fgets(buf, sizeof(buf), stream)))
      {
         if (dynamic)
            free(dynamic);
         return 0;
      }

      if (NULL == dynamic)
      {
         dynamic = (char *) malloc(strlen(buf) + 1);
         if (NULL == dynamic)
         {
            return 0;
         }
         strcpy(dynamic, buf);
      }
      else
      {
         /* a mini-version of realloc that works with our memory manager */
         char *temp = (char *) malloc(strlen(dynamic) + strlen(buf) + 1);
         if (NULL == temp)
            return 0;

         strcpy(temp, dynamic);
         free(dynamic);
         dynamic = temp;

         strcat(dynamic, buf);
      }

      if (feof(stream))
      {
         return dynamic;
      }
   } 
   while (dynamic[strlen(dynamic) - 1] != '\n');

   return dynamic;
}


config::config()
{
   mpVars = NULL;
   mSaveNeeded = false;
}


config::~config()
{
   my_cleanup(mpVars);
}


/* open loads from the disk the saved configuration. */
int config::open()
{
   FILE *config_file;

   if ((config_file = fopen(CONFIG_FILE, "r")))
   {
      char *line;
      char *group = NULL, *key = NULL, *value = NULL;

      mSaveNeeded = true;
      while ((line = my_getline(config_file)))
      {
         char *s;
         
         if ('\n' == line[strlen(line) - 1])
            line[strlen(line) - 1] = '\0';
         
         s = line;

         do 
         {
            /* eat up whitespace */
            while (isspace(*s))
               s++;

            switch (*s) 
            {
            case ';':
            case '#':
            case '\0':
               *s = '\0';
               break;

            case '[':
               if (group)
                  free(group);

               group = ++s;

               s = strchr(s, ']');
               if (NULL == s)
               {
                  std::cerr << "load_config: missing ']' after group\n";
                  s = group + strlen(group);
               }
               else
               {
                  *s++ = '\0';
               }

               if ((value = (char *) malloc(strlen(group) + 1)))
               {
                  strcpy(value, group);
               }
               group = value;
               break;

            default:
               key = s;
               s = strchr(s, '=');
               if (NULL == s)
               {
                  std::cerr << "load_config: missing '=' after key\n";
                  s = key + strlen(key);
               }
               else
               {
                  *s++ = '\0';
               }

               while (strlen(key) && isspace(key[strlen(key) - 1])) 
                  key[strlen(key) - 1] = '\0';

               while (isspace(*s)) 
                  s++;
               
               while (strlen(s) && isspace(s[strlen(s) - 1])) 
                  s[strlen(s) - 1]='\0';

               {
                  myvar_t *var = my_create(group ? group : "", key, s);
                  if (NULL == var)
                  {
                     std::cerr << "load_config: my_create failed\n";
                     return -1;
                  }

                  my_insert(&mpVars, var);
               }
               s += strlen(s);
            }
         } while (*s);

         free(line);
      }

      if (group) 
         free(group);

      fclose(config_file);
   }

   return 0;
}

/* save_config saves the current configuration to disk.*/
int config::close()
{
   FILE *config_file;
   char *group = "";

   if (!mSaveNeeded)
      return 0;

   config_file = fopen(CONFIG_FILE, "w");
   if (NULL == config_file)
   {
      std::cerr << "save_config failed\n";
      return -1;
   }

   fprintf(config_file, ";; " APPNAME " " VERSION "\n");
   fprintf(config_file, ";; NOTE: comments are not preserved.\n");
   my_save(config_file, mpVars, &group);
   fclose(config_file);

   return 0;
}


void config::write_int(const char *group, const char *key, int value)
{
   char buf[24];
   
   sprintf(buf, "%d", value);
   buf[sizeof(buf) - 1] = '\0';

   myvar_t *var = my_create(group, key, buf);
   if (NULL == var)
   {
      std::cerr << "write_int failed\n";
      return;
   }

   my_insert(&mpVars, var);
   mSaveNeeded = true;
}

/* read_int loads an integer from the configuration into "value"
**
** If the specified "key" does not exist, the "def"ault is returned
*/
int config::read_int(const char *group, const char *key, int def)
{
   myvar_t *var = my_lookup(mpVars, group, key);
   if (NULL == var)
   {
      write_int(group, key, def);
      
      return def;
   }

   return strtoul(var->value, 0, 0);
}


void config::write_string(const char *group, const char *key, const char *value)
{
   myvar_t *var = my_create(group, key, value);
   if (NULL == var)
   {
      std::cerr << "write_string failed\n";
      return;
   }

   my_insert(&mpVars, var);
   mSaveNeeded = true;
}

/* read_string copies a string from the configuration into "value"
**
** If the specified "key" does not exist, the "def"ault is returned
*/
const char *config::read_string(const char *group, const char *key, const char *def)
{
   myvar_t *var = my_lookup(mpVars, group, key);
   if (NULL == var)
   {
      if (def != NULL)
         write_string(group, key, def);

      return def;
   }

   return var->value;
}
