//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// system.cpp
//
// System-dependent routines
//
// $Id: $
//

#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include "system.h"
#include "assert.h"
#include "scanline.h"
#include "config.h"

#define  MIN(x, y)   (((x) < (y)) ? (x) : (y))

System::System()
{
   mpConfig = new config;
   if (mpConfig->open())
   {
      // we failed to open the config file
      delete mpConfig;
      mpConfig = NULL;
   }

   mSoundEnabled = false;
   mpAudioStream = NULL;
   mSampleRate = 11025;
   mAudioFragSize = 11025 / 60;
   mAudioPosition = 0;
   mAudioAvailable = mAudioFragSize;

   mpPixels = 0;
   mWidth = 0;
   mHeight = 0;
   mBPP = 0;
   mScale = 1;
   mPitch = 0;
   mPixelWidth = 0;
   mLoss[0] = mLoss[1] = mLoss[2] = 0;
   mPos[0] = mPos[1] = mPos[2] = 0;
   mIndexed = false;

   mVideoLock = 0;

   mTimer = 0;

   memset(mPalette4, 0, sizeof(mPalette4));
   memset(mPalette2, 0, sizeof(mPalette2));
   memset(mPalette1, 0, sizeof(mPalette1));
}


System::~System()
{
   ASSERT(0 == mVideoLock);

   if (mpConfig)
      mpConfig->close();
   delete mpConfig;
}


void System::audioCallback(unsigned char *pStream, int numSamples)
{
   if (!mSoundEnabled || !mAudioAvailable)
      return;

   int copyLength = MIN(mAudioAvailable, numSamples);
   if (mAudioPosition + copyLength > mAudioFragSize)
   {
      int chunk = mAudioFragSize - mAudioPosition;
      memcpy(pStream, mpAudioStream + mAudioPosition, chunk);
      memcpy(pStream + chunk, mpAudioStream, copyLength - chunk);
   }
   else
   {
      memcpy(pStream, mpAudioStream + mAudioPosition, copyLength);
   }

   mAudioPosition = (mAudioPosition + copyLength) % mAudioFragSize;
   mAudioAvailable -= copyLength;
}


void System::submitAudio(unsigned char *pStream, int numSamples)
{
   if (!mSoundEnabled || !mpAudioStream || !numSamples)
      return;

   while (mAudioFragSize - mAudioAvailable < numSamples)
      /* hang out */;

   disableInterrupts();

   int queuePosition = (mAudioPosition + mAudioAvailable) % mAudioFragSize;
   if (queuePosition + numSamples > mAudioFragSize)
   {
      int chunk = mAudioFragSize - queuePosition;
      memcpy(mpAudioStream + queuePosition, pStream, chunk);
      memcpy(mpAudioStream, pStream + chunk, numSamples - chunk);
   }
   else
   {
      memcpy(mpAudioStream + queuePosition, pStream, numSamples);
   }

   mAudioAvailable += numSamples;

   enableInterrupts();
}


void System::lockVideo()
{
   mVideoLock++;
}


void System::unlockVideo()
{
   mVideoLock--;
   ASSERT(mVideoLock >= 0);
}


void System::submitVideo(unsigned char *pBuffer, int scanline)
{
   switch (mScale)
   {
   case 1:
      if (scanline < mHeight)
      {
         unsigned char *pDest = mpPixels + scanline * mPitch;
         switch (mPixelWidth)
         {
         case 1:
            scan_expand_1(pDest, pBuffer, mPalette1, 256);
            break;

         case 2:
            scan_expand_2((uint16 *) pDest, pBuffer, mPalette2, 256);
            break;

         case 3:
            scan_expand_3(pDest, pBuffer, mPalette4, 256);
            break;

         case 4:
            scan_expand_4((uint32 *) pDest, pBuffer, mPalette4, 256);
            break;
         }
      }
      break;

   case 2:
      if (scanline < mHeight / 2)
      {
         unsigned char *pDest = mpPixels + scanline * mPitch * 2;
         switch (mPixelWidth)
         {
         case 1:
            scan_expand_2x_1(pDest, pBuffer, mPalette1, 256);
            break;

         case 2:
            scan_expand_2x_2((uint16 *) pDest, pBuffer, mPalette2, 256);
            break;

         case 3:
            scan_expand_2x_3(pDest, pBuffer, mPalette4, 256);
            break;

         case 4:
            scan_expand_2x_4((uint32 *) pDest, pBuffer, mPalette4, 256);
            break;
         }
         memcpy(pDest + mPitch, pDest, 256 * 2 * mPixelWidth);
      }
      break;
   }
}

void System::setColor(int index, int r, int g, int b, int value)
{
   if (-1 != value)
   {
      if (mIndexed)
      {
         mPalette1[index] = value;
         mPalette2[index] = (value << 8) | value;
         mPalette4[index] = (value << 24) | (value << 16) | (value << 8) | value;
      }
      else
      {
         r = (r >> mLoss[0]) << mPos[0];
         g = (g >> mLoss[1]) << mPos[1];
         b = (b >> mLoss[2]) << mPos[2];

         uint32 color = (r | g | b);

         switch (mPixelWidth)
         {
         case 1:
            mPalette1[index] = color;
            mPalette2[index] = (color << 8) | color;
            mPalette4[index] = (color << 24) | (color << 16) | (color << 8) | color;
            break;

         case 2:
            mPalette2[index] = color;
            mPalette4[index] = (color << 16) | color;
            break;

         case 3:
         case 4:
            mPalette4[index] = color;
            break;
         }
      }
   }
}


void System::silenceAudio()
{
   if (mSoundEnabled)
   {
      memset(mpAudioStream, 0x80, mAudioFragSize);
      mAudioPosition = 0;
      mAudioAvailable = mAudioFragSize;
   }
}


void System::mark()
{
   mTimer = (long long) clock() * 1000;
   //gettimeofday((timeval &) &mTimer, NULL);
}


int System::elapsed()
{
   uint32 now;
   int usecs;

   now = (uint32) clock() * 1000;
   usecs = now - mTimer;
   mTimer = now;

   return usecs;
   /*
   struct timeval tv;
   int secs, usecs;
   gettimeofday(&tv, NULL);
   secs = tv.tv_sec - ((timeval) mTimer).tv_sec;
   usecs = tv.tv_usec - ((timeval) mTimer).tv_usec;
   mTimer = (long long) tv;
   if (!secs)
      return usecs;
   return 1000000 + usecs;
   */
}


void System::usleep(int us)
{
   if (us > 0)
   {
      //sleep(us / 1000);
      usleep(us);
      /*
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = us;
      select(0, NULL, NULL, NULL, &tv);
      */
   }
}
   

int System::timerResolution() const
{
   return 1000;
}
